/******************************************************************************
FILE:				BoString.m
LANGUAGE:			Objective-C
SYSTEM:				ANSI (stdio,string)
USER-INTERFACE:		None
DESCRIPTION
	This class defines string object. The string data is dynamically allocated
	and reclaimed.
AUTHOR
	<PJB> Pascal J. Bourguignon
MODIFICATIONS
	1993/01/29 <PJB> Creation.
	1993/03/19 <PJB> Updated to use newly Bc- prefixed interfaces.
LEGAL
	Copyright Pascal J. Bourguignon 1993 - 1993
	All rights reserved.
	This program may not be included in any commercial product without the 
	author written permission. It may be used freely for any non-commercial 
	purpose, provided that this header is always included.
******************************************************************************/
#import <string.h>
#import <BoString.h>
#import <BcMem.h>
#import <BcImplementation.h>

@implementation BoString
/*
	char*			data
	INT32			length
	INT32			allocation
*/

/*
	INVARIANTS:
		data#NIL
		1<=allocation
		0<=length<allocation
		data[length]=(char)0
		for all i in [0..length-1], data[i]#(char)0
*/

#define AllocIncrement	(128)
#define Minimum(a,b)	(((a)<(b))?(a):(b))

	-(id)setCapacity:(INT32)nAllocation copy:(BOOLEAN)copy
	{
			char*		nData;
			INT32		nLength;
			
		if(nAllocation>1){
			nData=BcMem_Allocate(sizeof(char)*nAllocation);
			if(copy){
				nLength=Minimum(nAllocation-1,length);
				BcMem_Copy(data,nData,nLength*sizeof(char));
			}else{
				nLength=0;
			}
		}else{
			nAllocation=1;
			nData=BcMem_Allocate(sizeof(char)*nAllocation);
			nLength=0;
		}
		nData[nLength]=(char)0;
		BcMem_Deallocate((void**)(&data));
		data=nData;
		length=nLength;
		allocation=nAllocation;
		return(self);
	}//setCapacity;


// Object:

	-(id)init
	{
		[super init];
		allocation=1;
		data=BcMem_Allocate(sizeof(char)*allocation);
		length=0;
		data[length]=(char)0;
		return(self);
	}//init;


	-(id)initCapacity:(INT32)size
	{
		[self init];
		[self setCapacity:size copy:FALSE];
		return(self);
	}//initCapacity:;
	

	-(id)free
	{
		[self setCapacity:0 copy:FALSE];
		return([super free]);
	}//free;


// BoObject:


	-(void)printOn:(FILE*)file
	{
		[super printOn:file];
		fprintf(file,"data=           %p\n",(void*)data);
		if(data!=NIL){
			fprintf(file,"*data=          \"%s\"\n",data);
		}
		fprintf(file,"length=         %ld\n",length);
		fprintf(file,"allocation=     %ld\n",allocation);
	}//printOn:;
	
	
// BoString:

	-(INT32)length
	{
		return(length);
	}//length;
	
	
	-(const char*)string
	{
		return(data);
	}//string;


	-(id)setString:(const char*)nString
	{
			INT32		nLength;
		
		nLength=strlen(nString);
		if(nLength>=allocation){
			[self setCapacity:nLength+1 copy:FALSE];
		}
		BcMem_Copy((void*)nString,data,nLength*sizeof(char));
		data[nLength]=(char)0;
		length=nLength;
		return(self);
	}//setString:;


	-(id)getString:(char*)string
	{
		BcMem_Copy(data,string,length*sizeof(char));
		string[length]=(char)0;
		return(self);
	}//getString:;


	-(char)charAt:(INT32)index
	{
		if((index<0)OR(index>=length)){
			return((char)0);
		}else{
			return(data[index]);
		}
	}//charAt:;


	-(id)setChar:(char)nChar at:(INT32)index
	{
		if((0<=index)AND(index<=length)){
			if(nChar==(char)0){
				length=index;
			}else if(index==length){
				if(allocation<=length+1){
					[self setCapacity:allocation+AllocIncrement copy:TRUE];
				}
				INC(length);
				data[length]=(char)0;
			}
			data[index]=nChar;			
		}
		return(self);
	}//setChar:;


	-(id)insert:(BoString*)substring at:(INT32)index
	{
			INT32		nLength;

		if(index<0){
			index=0;
		}else if(length<index){
			index=length;
		}
		nLength=length+substring->length;
		if(nLength>=allocation){
			// SEE: We copy some char of self twice !
			[self setCapacity:nLength+1 copy:TRUE];
		}
		BcMem_Copy(data+index,data+index+substring->length,
						(length-index+1)*sizeof(char));
		BcMem_Copy(substring->data,data+index,substring->length*sizeof(char));
		length=nLength;
		return(self);
	}//insert:at:;


	-(id)deleteFrom:(INT32)index length:(INT32)delen
	{
		if(index<0){
			index=0;
		}else if(length<=index){
			return(self);
		}
		if(length<index+delen){
			delen=length-index;
		}
		BcMem_Copy(data+index+delen,data+index,
					(length-(index+delen)+1)*sizeof(char));
		length=length-delen;
		return(self);
	}//deleteFrom:length:;


	-(INT32)position:(BoString*)substring from:(INT32)index
	{
		// SEE: could we use the FastStringSearch algorithm?
			INT32			pos;
			INT32			i;
			INT32			max;
		
		max=length-substring->length;		
		if(index<0){
			pos=0;
		}else{
			pos=index;
		}
		while(pos<=max){
			i=0;
			while(data[pos+i]==substring->data[i]){
				INC(i);
			}
			if(substring->data[i]==(char)0){
				return(pos);
			}
			INC(pos);
		}
		// not found
		return(-1);
	}//position:;


	-(id)copy:(BoString*)string from:(INT32)index length:(INT32)copylen
	{
		if(index+copylen>string->length){
			copylen=string->length-index;
		}
		if(copylen<=0){
			data[0]=(char)0;
			length=0;
			return(self);
		}
		if(copylen>=allocation){
			[self setCapacity:copylen+1 copy:FALSE];
		}
		BcMem_Copy(string->data+index,data,copylen*sizeof(char));
		data[copylen]=(char)0;
		length=copylen;
		return(self);
	}//copy:from:length:;


	-(id)append:(BoString*)tail
	{
		return([self insert:tail at:length]);
	}//append:;


	-(int)compare:(BoString*)other
	{
		return(strcmp([self string],[other string]));	
	}//compare:;
	
@end //BoString
